from __future__ import annotations

import mlx.core as mx

from src.metrics.metric import Metric


class Accuracy(Metric):
    def __init__(self):
        super().__init__("accuracy")
        self.add_state("correct", 0, reduce_fx=None)
        self.add_state("total", 0, reduce_fx=None)

    def update(self, predictions, targets):
        self.correct += (mx.argmax(predictions, axis=1) == targets).sum().item()
        self.total += targets.shape[0]

    def compute(self):
        return self.correct / self.total if self.total > 0 else 0.0



from __future__ import annotations

import mlx.core as mx
import numpy as np

from src.core.datamodule import DataModule
from src.core.trainmodule import TrainModule
from src.loops.loop import Loop
from src.metrics import Accumulation
from src.metrics.metric import Metric


class ValidationLoop(Loop):
    def __init__(self, trainer, train_module: TrainModule, data_module: DataModule) -> None:
        super().__init__(trainer, train_module, data_module)
        self.metrics: dict[str, Metric] = {"loss": Accumulation()}

    def on_validation_epoch_start(self, *args, **kwargs):
        """Validation hook which triggers at the beginning of a validation epoch.

        - Sets the model to validation mode.
        - Calls user-defined methods and resets metrics.
        """
        self.model.train(False)
        self._call_user_method("on_validation_epoch_start", args, kwargs)
        self._reset_metrics()

    def on_validation_batch_start(self, batch: np.array, batch_idx: int) -> mx.array:
        """Validation hook which triggers at the start of a validation batch.

        - Converts the batch to mx.array format.
        - Calls user-defined methods.

        Args:
            batch : A batch of data.
            batch_idx: Index of the batch.

        Returns:
            batch: Batch formatted as mx.array with additional user-defined transformations.
        """
        batch = self._call_user_method("on_validation_batch_start", batch, batch_idx)
        batch = {k: mx.array(v) for k, v in batch.items()}
        return batch

    def on_validation_batch_end(self, loss):
        """Validation hook which triggers at the end of a validation batch.

        - Calls user-defined methods.
        - Updates the loss metric.
        """
        self._call_user_method("on_validation_batch_end", loss)
        self.metrics["loss"].update(loss)

    def validation_step(self, batch, batch_idx) -> tuple[mx.array, mx.array]:
        """Executes a validation step, computing loss.

        Args:
            batch: A batch of data.
            batch_idx: Index of the batch.

        Returns:
            Tuple containing the loss.
        """
        assert hasattr(self.train_module, "validation_step"), "The train module has no validation_step defined"
        loss = self.train_module.validation_step(batch, batch_idx)
        return loss

    def on_validation_epoch_end(self, *args, **kwargs):
        """Validation hook which triggers at the end of a validation epoch.

        - Calls user-defined methods.
        - Resets the data loader.
        - Logs the validation metrics.
        """
        self._call_user_method("on_validation_epoch_end", args, kwargs)
        self.data_module.valid_dataloader().reset()
        self._log_epoch_metrics()

    def iterate(self):
        """Main loop iteration for validation.

        - Triggers events and hooks at the start and end of each epoch and batch.
        - Executes the validation step for each batch.
        """
        epoch_events = {
            "start_event": "on_validation_epoch_start",
            "end_event": "on_validation_epoch_end",
            "start_method": self.on_validation_epoch_start,
            "end_method": self.on_validation_epoch_end,
        }

        batch_events = {
            "start_event": "on_validation_batch_start",
            "end_event": "on_validation_batch_end",
            "step_event": "validation_step",
            "start_method": self.on_validation_batch_start,
            "end_method": self.on_validation_batch_end,
        }

        self._execute_with_events(
            epoch_events=epoch_events,
            batch_events=batch_events,
            batch_iterator=self.data_module.valid_dataloader(),
            step_method=self.validation_step,
        )

    def _log_epoch_metrics(self):
        """Logs the metrics at the end of an epoch."""
        print(
            " | ".join(
                (
                    f"Epoch {self.current_epoch:02d}",
                    f"Validation loss: {self.metrics['loss'].compute().item():.3f}",
                    f"Validation accuracy: {self.metrics['validation_accuracy'].compute():.3f}",
                )
            )
        )



"""
Implementation of ResNets for CIFAR-10 as per the original paper [https://arxiv.org/abs/1512.03385].
Configurations include ResNet-20, ResNet-32, ResNet-44, ResNet-56, ResNet-110, ResNet-1202.
"""
from __future__ import annotations

import mlx.core as mx
from mlx import nn
from mlx.utils import tree_flatten

__all__ = [
    "ResNet",
    "resnet20",
    "resnet32",
    "resnet44",
    "resnet56",
    "resnet110",
    "resnet1202",
]


class ShortcutA(nn.Module):
    def __init__(self, dims):
        super().__init__()
        self.dims = dims

    def __call__(self, x):
        return mx.pad(
            x[:, ::2, ::2, :],
            pad_width=[(0, 0), (0, 0), (0, 0), (self.dims // 4, self.dims // 4)],
        )


class Block(nn.Module):
    """Implements a ResNet block with two convolutional layers and a skip connection.

    As per the paper, CIFAR-10 uses Shortcut type-A skip connections. (See paper for details)
    """

    def __init__(self, in_dims, dims, stride=1):
        super().__init__()

        self.conv1 = nn.Conv2d(in_dims, dims, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm(dims)

        self.conv2 = nn.Conv2d(dims, dims, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm(dims)

        if stride != 1:
            self.shortcut = ShortcutA(dims)
        else:
            self.shortcut = None

    def __call__(self, x):
        out = nn.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        if self.shortcut is None:
            out += x
        else:
            out += self.shortcut(x)
        out = nn.relu(out)
        return out


class ResNet(nn.Module):
    """Creates a ResNet model for CIFAR-10, as specified in the original paper."""

    def __init__(self, block, num_blocks, num_classes=10):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm(16)

        self.layer1 = self._make_layer(block, 16, 16, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 16, 32, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 32, 64, num_blocks[2], stride=2)

        self.linear = nn.Linear(64, num_classes)

    def _make_layer(self, block, in_dims, dims, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(in_dims, dims, stride))
            in_dims = dims
        return nn.Sequential(*layers)

    def num_params(self):
        nparams = sum(x.size for k, x in tree_flatten(self.parameters()))
        return nparams

    def __call__(self, x):
        x = nn.relu(self.bn1(self.conv1(x)))
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = mx.mean(x, axis=[1, 2]).reshape(x.shape[0], -1)
        x = self.linear(x)
        return x


def resnet20(**kwargs):
    return ResNet(Block, [3, 3, 3], **kwargs)


def resnet32(**kwargs):
    return ResNet(Block, [5, 5, 5], **kwargs)


def resnet44(**kwargs):
    return ResNet(Block, [7, 7, 7], **kwargs)


def resnet56(**kwargs):
    return ResNet(Block, [9, 9, 9], **kwargs)


def resnet110(**kwargs):
    return ResNet(Block, [18, 18, 18], **kwargs)


def resnet1202(**kwargs):
    return ResNet(Block, [200, 200, 200], **kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod
from types import SimpleNamespace
from typing import Any

from mlx import nn


class TrainModule(ABC):
    def __init__(self, args: SimpleNamespace) -> None:
        super().__init__()
        self.args = args

    @property
    def trainer(self):
        return self._trainer

    @trainer.setter
    def trainer(self, trainer):
        self._trainer = trainer

    @property
    def model(self):
        return self._model

    @model.getter
    def model(self) -> nn.Module:
        if not hasattr(self, "_model"):
            raise ValueError("Model not defined, please set the model in your trainmodule.")
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    def log(self, name, value):
        self.trainer.log(name, value)

    def forward(self, *args: Any, **kwargs: Any) -> Any:
        r"""Same as :meth:`torch.nn.Module.forward`.

        Args:
            *args: Whatever you decide to pass into the forward method.
            **kwargs: Keyword arguments are also possible.

        Return:
            Your model's output
        """
        return super().forward(*args, **kwargs)

    @abstractmethod
    def training_step(self, batch: dict, batch_index: int, *args, **kwargs):
        """Here you compute and return the training loss and some additional metrics for e.g. the
        progress bar or logger.

        Args:
            batch: The output of your data iterable, normally a :class:`~torch.utils.data.DataLoader`.

        Return:
            - :class:`~mx.Array` - The loss tensor
            - ``dict`` - A dictionary which can include any keys, but must include the key ``'loss'`` in the case of
              automatic optimization.

        In this step you'd normally do the forward pass and calculate the loss for a batch.
        """
        pass

    @abstractmethod
    def configure_optimizers(self):
        r"""Choose what optimizers to use in your optimization.

        Normally you'd need one. But in the case of GANs or similar you might have multiple.
        """
        pass



from __future__ import annotations

import mlx.core as mx
import numpy as np
from mlx import nn

from src.core.datamodule import DataModule
from src.core.trainmodule import TrainModule
from src.loops.loop import Loop
from src.metrics import Accumulation
from src.metrics.metric import Metric


class TrainLoop(Loop):
    def __init__(self, trainer, train_module: TrainModule, data_module: DataModule) -> None:
        super().__init__(trainer, train_module, data_module)
        self.metrics: dict[str, Metric] = {"loss": Accumulation()}

    def on_train_epoch_start(self, *args, **kwargs):
        """Training hook which triggers at the beginning of a training epoch.

        - Sets the model to training mode.
        - Initializes the state.
        - Calls user-defined methods and resets metrics.
        """
        self.model.train(True)
        self.state = [self.model.state, self.trainer.optimizer.state]
        self._call_user_method("on_train_epoch_start", args, kwargs)
        self._reset_metrics()

    def on_train_batch_start(self, batch: np.array, batch_idx: int) -> mx.array:
        """Training hook which triggers at the start of a training batch.

        - Converts the batch to mx.array format.
        - Calls user-defined methods.

        Args:
            batch : A batch of data.
            batch_idx: Index of the batch.

        Returns:
            batch: Batch formatted as mx.array with additional user-defined transformations.
        """
        batch = self._call_user_method("on_train_batch_start", batch, batch_idx)
        batch = {k: mx.array(v) for k, v in batch.items()}
        return batch

    def on_train_batch_end(self, loss):
        """Training hook which triggers at the end of a training batch.

        - Calls user-defined methods.
        - Evaluates the model state and updates metrics.
        """
        self._call_user_method("on_train_batch_end", loss)
        mx.eval(self.state)
        self.metrics["loss"].update(loss)

    def training_step(self, batch, batch_idx) -> tuple[mx.array, mx.array]:
        """Executes a training step, computing loss and gradients.

        Args:
            batch: A batch of data.
            batch_idx: Index of the batch.

        Returns:
            Tuple containing the loss and gradients.
        """
        assert hasattr(self.train_module, "training_step"), "The train module has no training_step defined"
        loss, grads = nn.value_and_grad(self.model, self.train_module.training_step)(batch, batch_idx)
        self.trainer.optimizer.update(self.model, grads)
        return loss

    def on_train_epoch_end(self, *args, **kwargs):
        """Training hook which triggers at the end of a training epoch.

        - Calls user-defined methods.
        - Resets the data loader.
        - Logs the training metrics.
        """
        self._call_user_method("on_train_epoch_end", args, kwargs)
        self.data_module.train_dataloader().reset()
        self._log_epoch_metrics()

    def iterate(self):
        """Main loop iteration for training.

        - Triggers events and hooks at the start and end of each epoch and batch.
        - Executes the training step for each batch.
        """
        epoch_events = {
            "start_event": "on_train_epoch_start",
            "end_event": "on_train_epoch_end",
            "start_method": self.on_train_epoch_start,
            "end_method": self.on_train_epoch_end,
        }

        batch_events = {
            "start_event": "on_train_batch_start",
            "end_event": "on_train_batch_end",
            "step_event": "training_step",
            "start_method": self.on_train_batch_start,
            "end_method": self.on_train_batch_end,
        }

        self._execute_with_events(
            epoch_events=epoch_events,
            batch_events=batch_events,
            batch_iterator=self.data_module.train_dataloader(),
            step_method=self.training_step,
        )

    def _log_epoch_metrics(self):
        """Logs the metrics at the end of an epoch."""
        print(
            " | ".join(
                (
                    f"Epoch {self.current_epoch:02d}",
                    f"Training loss: {self.metrics['loss'].compute().item():.3f}",
                    f"Training accuracy: {self.metrics['train_accuracy'].compute():.3f}",
                )
            )
        )



from __future__ import annotations

from src.callbacks.callback import Callback


class ModelSummary(Callback):
    def __init__(self) -> None:
        super().__init__()

    def on_fit_start(self, trainer):
        print(trainer.model)



from __future__ import annotations

from types import SimpleNamespace
from typing import Any

import mlx.optimizers as optim
from mlx import nn

from examples.cifar_datamodule import Cifar10DataModule
from examples.resnet import resnet20
from src.core.trainer import Trainer
from src.core.trainmodule import TrainModule
from src.metrics import Accuracy


class CifarTrainModule(TrainModule):
    def __init__(self, args: SimpleNamespace) -> None:
        super().__init__(args)
        self.model = resnet20()

    def configure_optimizers(self):
        return optim.Adam(learning_rate=1e-3)

    def forward(self, x) -> Any:
        return self.model(x)

    def setup(self):
        self.accs = []
        self.validation_accuracy = Accuracy()
        self.train_accuracy = Accuracy()

    def training_step(self, batch: dict, batch_idx):
        x, y = batch["image"], batch["label"]
        y_hat = self.forward(x)
        loss = nn.losses.cross_entropy(y_hat, y, reduction="mean")
        self.train_accuracy(y_hat, y)
        self.log("train_accuracy", self.train_accuracy)
        return loss

    def validation_step(self, batch: dict, batch_idx):
        x, y = batch["image"], batch["label"]
        y_hat = self.forward(x)
        loss = nn.losses.cross_entropy(y_hat, y, reduction="mean")
        self.validation_accuracy(y_hat, y)
        self.log("validation_accuracy", self.validation_accuracy)
        return loss


if __name__ == "__main__":
    datamodule = Cifar10DataModule(args=SimpleNamespace(batch_size=256))
    trainmodule = CifarTrainModule(args=None)
    trainer = Trainer(train_module=trainmodule, data_module=datamodule, max_epochs=30, run_validation_every_n_epochs=1)
    trainer.fit()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any

import mlx.core as mx


class Metric(ABC):
    def __init__(self, name) -> None:
        super().__init__()
        self.name = name
        self._allowed_reduce_fx = [mx.mean, mx.sum, mx.concatenate]
        self.reduce_fx: Callable = mx.mean
        self._defaults: dict[str, Any] = {}

    def add_state(self, name: str, default: mx.array, reduce_fx: Callable) -> None:
        if hasattr(self, name):
            raise ValueError(f"State {name} already defined")
        if reduce_fx is not None and reduce_fx.__name__ not in [m.__name__ for m in self._allowed_reduce_fx]:
            raise ValueError(f"Reduction function {reduce_fx.__name__} not supported.")
        self.reduce_fx = reduce_fx
        self._defaults[name] = default
        setattr(self, name, default)

    def __call__(self, *args, **kwargs) -> None:
        self.update(*args, **kwargs)

    def reset(self):
        for state_variable_name, default_value in self._defaults.items():
            setattr(self, state_variable_name, default_value)

    @abstractmethod
    def update(self, *args, **kwargs):
        pass

    @abstractmethod
    def compute(self):
        pass



from __future__ import annotations

from abc import ABC, abstractmethod


class DataModule(ABC):
    @abstractmethod
    def setup(self) -> None:
        pass

    @abstractmethod
    def train_dataloader(self):
        pass

    @abstractmethod
    def valid_dataloader(self):
        pass

    @abstractmethod
    def test_dataloader(self):
        pass



from __future__ import annotations

from types import SimpleNamespace

import numpy as np
from mlx.data.datasets.cifar import load_cifar10

from src.core.datamodule import DataModule


class Cifar10DataModule(DataModule):
    mean = np.array([0.485, 0.456, 0.406]).reshape((1, 1, 3))
    std = np.array([0.229, 0.224, 0.225]).reshape((1, 1, 3))

    @staticmethod
    def normalize(x):
        x = x.astype("float32") / 255.0
        return (x - Cifar10DataModule.mean) / Cifar10DataModule.std

    def __init__(self, args: SimpleNamespace) -> None:
        super().__init__()
        self.args = args

    def setup(self) -> None:
        self.train_dataset = load_cifar10(root=None, train=True)
        self.valid_dataset = load_cifar10(root=None, train=False)

    def train_dataloader(self):
        if not hasattr(self, "_train_dataloader"):
            self._train_dataloader = (
                self.train_dataset.shuffle()
                .to_stream()
                .image_random_h_flip("image", prob=0.5)
                .pad("image", 0, 4, 4, 0.0)
                .pad("image", 1, 4, 4, 0.0)
                .image_random_crop("image", 32, 32)
                .key_transform("image", Cifar10DataModule.normalize)
                .batch(self.args.batch_size)
                .prefetch(4, 4)
            )
        return self._train_dataloader

    def valid_dataloader(self):
        if not hasattr(self, "_valid_dataloader"):
            self._valid_dataloader = (
                self.valid_dataset.to_stream()
                .key_transform("image", Cifar10DataModule.normalize)
                .batch(self.args.batch_size)
            )
        return self._valid_dataloader

    def test_dataloader(self):
        return self.valid_dataloader()



from __future__ import annotations

import enum
from abc import ABC, abstractmethod

from mlx import nn

from src.core.datamodule import DataModule
from src.core.trainmodule import TrainModule


class LoopType(enum.Enum):
    TRAINING = enum.auto()
    VALIDATION = enum.auto()


class Loop(ABC):
    """Base class for loops."""

    def __init__(self, trainer, train_module: TrainModule, data_module: DataModule) -> None:
        super().__init__()
        self.train_module = train_module
        self.data_module = data_module
        self.trainer = trainer

    def setup(self):
        """Method which makes sure both the train module and the data modules are properly
        setup."""
        self.data_module.setup()
        self.train_module.setup()

    def log(self, name, value):
        self.metrics[name] = value

    # Helper methods
    def _call_user_method(self, method_name, *args, **kwargs):
        """Calls a user-defined method if it exists."""
        if hasattr(self.train_module, method_name):
            return getattr(self.train_module, method_name)(*args, **kwargs)
        return args[0] if args else None

    def _reset_metrics(self):
        """Resets all metrics at the start of an epoch."""
        for metric in self.metrics.values():
            metric.reset()

    @property
    def model(self) -> nn.Module:
        return self.train_module.model

    @property
    def current_epoch(self) -> int:
        return self._current_epoch

    @current_epoch.setter
    def current_epoch(self, current_epoch) -> None:
        self._current_epoch = current_epoch

    def _execute_with_events(self, epoch_events, batch_events, batch_iterator, step_method):
        """Helper method to execute loop iterations with events.

        Args:
            epoch_events (dict): Dictionary containing start and end events and methods for the epoch.
                {
                    "start_event": str,
                    "end_event": str,
                    "start_method": callable,
                    "end_method": callable
                }
            batch_events (dict): Dictionary containing start, end, and step events and methods for the batch.
                {
                    "start_event": str,
                    "end_event": str,
                    "step_event": str,
                    "start_method": callable,
                    "end_method": callable
                }
            batch_iterator (iterable): Iterator for the batches.
            step_method (callable): Method to perform the step operation.
        """
        # Trigger start of epoch events and methods
        self.trainer._trigger_event(epoch_events["start_event"])
        epoch_events["start_method"]()

        for batch_idx, batch in enumerate(batch_iterator):
            # Trigger start of batch events and methods
            self.trainer._trigger_event(batch_events["start_event"])
            batch = batch_events["start_method"](batch, batch_idx)

            # Perform the step operation
            loss = step_method(batch, batch_idx)
            self.trainer._trigger_event(batch_events["step_event"])

            # Trigger end of batch events and methods
            batch_events["end_method"](loss)
            self.trainer._trigger_event(batch_events["end_event"])

        # Trigger end of epoch events and methods
        epoch_events["end_method"]()
        self.trainer._trigger_event(epoch_events["end_event"])

    @abstractmethod
    def iterate(self):
        raise NotImplementedError



from __future__ import annotations

from contextlib import contextmanager

from mlx import nn

from src.callbacks.model_summary import ModelSummary
from src.core.datamodule import DataModule
from src.core.trainmodule import TrainModule
from src.loops.loop import LoopType
from src.loops.train_loop import TrainLoop
from src.loops.validation_loop import ValidationLoop


class Trainer:
    def __init__(
        self,
        train_module: TrainModule,
        data_module: DataModule,
        max_epochs: int,
        run_validation_every_n_epochs: int = 1,
        run_sanity_validation: bool = True,
        **kwargs,
    ) -> None:
        self.train_module = train_module
        self.train_module.trainer = self  # Set reference to the trainer inside the train module
        self.data_module = data_module
        self.max_epochs = max_epochs
        self.run_validation_every_n_epochs = run_validation_every_n_epochs
        self.run_sanity_validation = run_sanity_validation

        # Initialize model and optimizer
        self.model = train_module.model
        self.optimizer = self.train_module.configure_optimizers()

        # Initialize callbacks
        self.callbacks = [ModelSummary()]

        # Initialize loops
        self.loops = {
            LoopType.VALIDATION: ValidationLoop(self, train_module, data_module),
            LoopType.TRAINING: TrainLoop(self, train_module, data_module),
        }

    # Property for model
    @property
    def model(self) -> nn.Module:
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    # Property for current epoch
    @property
    def current_epoch(self) -> int:
        if not hasattr(self, "_current_epoch"):
            return 0
        return self._current_epoch

    @current_epoch.setter
    def current_epoch(self, current_epoch) -> None:
        self._current_epoch = current_epoch
        # Set the current epoch in each loop
        for loop in self.loops.values():
            loop.current_epoch = current_epoch

    # Property for active loop
    @property
    def active_loop(self) -> LoopType:
        return self._active_loop

    @active_loop.setter
    def active_loop(self, active_loop: LoopType):
        self._active_loop = active_loop

    # Property for optimizer
    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer):
        self._optimizer = optimizer

    @contextmanager
    def loop_context(self, loop_type: LoopType):
        """Context manager to set the active loop type."""
        self.active_loop = loop_type
        try:
            yield
        finally:
            self.active_loop = LoopType.TRAINING

    def log(self, name, value):
        """Logs a value during the active loop."""
        active_loop = self.loops[self.active_loop]
        active_loop.log(name, value)

    def register_callback(self, callback):
        """Registers a new callback."""
        self.callbacks.append(callback)

    def _trigger_event(self, event):
        """Triggers a specific event for all registered callbacks."""
        for callback in self.callbacks:
            if hasattr(callback, event):
                getattr(callback, event)(self)

    def fit(self):
        """Runs the fit process for the given number of epochs.

        Handles setup, sanity validation, training, and validation loops.
        """
        self._setup()
        self._trigger_event("on_fit_start")

        for epoch_number in range(self.max_epochs):
            self.current_epoch = epoch_number

            if self._should_run_sanity_validation():
                self._run_sanity_validation()

            self._run_training_loop()

            if self._should_run_validation():
                self._run_validation_loop()

    def _setup(self):
        """Configures the data and training modules."""
        self.data_module.setup()
        self.train_module.setup()

    def _should_run_sanity_validation(self):
        """Determines if sanity validation should run."""
        return self.run_sanity_validation and self.current_epoch == 0

    def _run_sanity_validation(self):
        """Runs the sanity validation loop."""
        with self.loop_context(LoopType.VALIDATION):
            self.loops[LoopType.VALIDATION].iterate()

    def _run_training_loop(self):
        """Runs the training loop for the current epoch."""
        with self.loop_context(LoopType.TRAINING):
            self.loops[LoopType.TRAINING].iterate()

    def _should_run_validation(self):
        """Determines if validation should run this epoch."""
        return self.current_epoch % self.run_validation_every_n_epochs == 0

    def _run_validation_loop(self):
        """Runs the validation loop for the current epoch."""
        with self.loop_context(LoopType.VALIDATION):
            self.loops[LoopType.VALIDATION].iterate()



from __future__ import annotations

import mlx.core as mx

from src.metrics.metric import Metric


class Accumulation(Metric):
    def __init__(self) -> None:
        super().__init__("accumulation")
        self.add_state("quantity", default=mx.array([]), reduce_fx=mx.mean)

    def update(self, quantity: mx.array):
        self.quantity = mx.concatenate([self.quantity, mx.expand_dims(quantity, 0)])

    def compute(self) -> mx.array:
        return self.reduce_fx(self.quantity)



from __future__ import annotations


class Callback:
    def on_fit_start(self, trainer):
        pass

    def on_fit_end(self, trainer):
        pass

    def on_train_epoch_start(self, trainer):
        pass

    def on_train_epoch_end(self, trainer):
        pass

    def on_train_batch_start(self, trainer):
        pass

    def training_step(self, trainer):
        pass

    def on_train_batch_end(self, trainer):
        pass

    def on_validation_epoch_start(self, trainer):
        pass

    def on_validation_step_start(self, trainer):
        pass

    def on_validation_step_end(self, trainer):
        pass

    def on_validation_epoch_end(self, trainer):
        pass
